from decorator import decorator
from ...api.store.client import handshake
from prompt_toolkit import print_formatted_text, HTML

HOST = None
PORT = None
INV_PROMPT = ">>>"

def getPrompt():
    return INV_PROMPT

@decorator
def bConnect(f, *args, **kwargs):
    if HOST is None:
        print_formatted_text(HTML(f"<ansired>Please <u>connect</u> first</ansired>"))
        return
    return f(*args, **kwargs)

def connect(*args):
    global HOST, PORT, INV_PROMPT
    if len(args) != 2:
        print_formatted_text(HTML(f"<ansired><u>connect</u> expects 2 arguments</ansired>"))
        return False
    try : 
        _ = handshake(*args)
        HOST=args[0]
        PORT=args[1]
        print_formatted_text(HTML(f"<ansigreen>Connection successfull at {HOST}:{PORT}</ansigreen>"))
        INV_PROMPT = f"{HOST}:{PORT}>"
    except :
        print_formatted_text(HTML(f"<ansired>Connection failed at {args[0]}:{args[1]} </ansired>"))


from prompt_toolkit import PromptSession, print_formatted_text, HTML
from prompt_toolkit.key_binding import KeyBindings
from .helpers import completer, customAutoSuggest
import re
from .commands import getPrompt, Executor


executor = Executor()

def digest(_input):
    if re.match('^[\s]*$', _input) :
        #print("Empty string")
        return
    _ = _input.split()
    #print("input::", _)
    
    cmd = _.pop(0)
    if cmd == 'help':
        _help(_[0] if _ else None)
        return True
    if not executor.isa(cmd):
        print_formatted_text(HTML(f"<ansired><b>{cmd}</b> is not a valid command</ansired>\nType <u>help</u> for a list of available commands"))
        return False

    _ = executor.process(cmd, *_)

def _help(cmd):
    if cmd is None:
        msg = ', '.join([ f"<u>{c}</u>" for c in executor.availbleCmd ])
        msg = f"Available command: {msg}"
        print_formatted_text(HTML(msg))
    elif not executor.isa(cmd):
        print_formatted_text(HTML(f"<ansired><b>{cmd}</b> is not a valid command</ansired>\nType <u>help</u> for a list of available commands"))
    else:
        msg = executor.help(cmd)
        #print("##", msg, "##")
        print_formatted_text(HTML(f"<u>{cmd}</u>:\n{msg}"))


kb = KeyBindings()
@kb.add('c-c')
def exit_(event):
    """
    Pressing Ctrl-Q will exit the user interface.

    """
    #_exit()
    executor.process('exit')


def run():
    session = PromptSession()
    
    while True:
        answer = session.prompt(getPrompt(), completer=completer, auto_suggest=customAutoSuggest(),
        key_bindings=kb )
        digest(answer)
        #print('You said: %s' % answer)
    exit(1)



UNIVERSAL_TREES = {}

UNIVERSAL_VECTORS = {}

def clear():
    global UNIVERSAL_TREES, UNIVERSAL_VECTORS
    print(f"Clearing local stores content")

    UNIVERSAL_VECTORS = {}
    UNIVERSAL_TREES   = {}

def delTreeByTaxids(*taxids):
    miss = []
    for taxid in taxids:
        if not taxid in UNIVERSAL_TREES:
            miss.append(taxid)
        else:
            UNIVERSAL_TREES.pop(taxid)
    if miss:
        raise KeyError(f"{miss} to delete element tree not found in local store")

def delVectorsByTaxid(*taxids):
    miss = []
    for taxid in taxids:
        if not taxid in UNIVERSAL_VECTORS:
            miss.append(taxid)
        else:
            UNIVERSAL_VECTORS.pop(taxid)
    
    print(f"Following vector element were deleted {set(taxids) - set(miss)}")
    if miss:
        raise KeyError(f"{miss} to delete element vector not found in local store")

def storeTreeByTaxid(tree, taxid):
    print(f"localTreeStoring -->{taxid}")

    global UNIVERSAL_TREES
    if taxid in UNIVERSAL_TREES:
        raise KeyError(f"{taxid} tree already exists in local store")
        
    UNIVERSAL_TREES[taxid] = tree

def getUniversalTree(fullTreeKey):
    if not fullTreeKey in UNIVERSAL_TREES:
        raise KeyError(f"No tree names {fullTreeKey} found in local store")
    
    return UNIVERSAL_TREES[fullTreeKey]

def storeVector(vector, fullVectorKey):
    print(f"localVectorStoring -->{fullVectorKey}")

    global UNIVERSAL_VECTORS
    if fullVectorKey in UNIVERSAL_VECTORS:
        raise KeyError(f"{fullVectorKey} vector already exists in local store")
        
    UNIVERSAL_VECTORS[taxid] = tree

def getUniversalVector(taxid):
    if not taxid in UNIVERSAL_VECTORS:
        raise KeyError(f"No taxid {taxid} found in vector local store")
    
    return UNIVERSAL_VECTORS[taxid]

def getTaxidKeys():
    return list(UNIVERSAL_TREES.keys())

def listTreeKey(*args,**kwargs):
    return list(UNIVERSAL_TREES.keys())

def listVectorKey(*args,**kwargs):
    return list(UNIVERSAL_VECTORS.keys())



from flask import Flask, jsonify, abort, request, make_response, Response
from ..data_objects import CulledGoParametersSchema as goParameterValidator
from ..data_objects import loadUnivGO
from .cache import setCacheType, delTreeByTaxids, storeTreeByTaxid, getTaxidKeys, getUniversalTrees
from .cache import getCulledVectors, storeCulledVector, getUniversalVectors, delVectorsByTaxid
from .cache import clear as clearStore, deleteTaxids
from .cache import buildUniversalVector, listTrees, listVectors, listMissUniversalVector, listCulled
from .cache import status as storeStatus
from decorator import decorator
from ...tree import enumNSkeysRevert as humanReadableNS
# IF OFFLOADABLE BLOCKING ROUTES KEEP ON SHOWING UP
# https://blog.miguelgrinberg.com/post/using-celery-with-flask

import time

from multiprocessing import Process, Value, Semaphore

bSemaphore = None
C_TYPE = None
_MAIN_ = False

def bootstrap(newElem=None, cacheType='local',\
    clear=False, _main_=False, **kwargs):
    global _MAIN_, C_TYPE
    
    if _main_:
        print("Bootstraping main process")
        _MAIN_ = _main_
        global bSemaphore
        bSemaphore = Semaphore(1) # bleeding eyes
        setCacheType(cacheType, **kwargs)
        C_TYPE = cacheType
        if clear:
            clearStore()

        if newElem:
            print("Boostraping unigo store with following new ressources")
            for taxid, ns, tree in newElem:
                print(f"\tAdding{taxid}:{tree}")
                storeTreeByTaxid(tree, taxid)

        tcount, vcount, _vcount = storeStatus()
        print(f"Database content:\n\t{tcount} trees, {vcount} vectors, {_vcount} culled")

        app = Flask(__name__)

        app.add_url_rule('/ping', 'ping', ping)
        
        app.add_url_rule('/taxids', 'view_taxids', view_taxids)
        
        app.add_url_rule('/unigos/<taxid>', 'view_unigos', view_unigos, methods=['GET'])
        
        app.add_url_rule('/vectors/<taxid>', 'view_vectors', view_vectors, methods=['GET'])
        
        app.add_url_rule("/vectors/<taxid>", 'view_culled_vectors', view_culled_vectors, methods=['POST'])
        
        app.add_url_rule('/add/taxid/<taxid>', 'add_unigo3NS', add_unigo3NS, methods=['POST'])

        app.add_url_rule('/del/taxid/<taxid>', 'del_taxonymy', del_taxonymy, methods=['DELETE'])

        app.add_url_rule('/add/unigo/<taxid>', 'add_unigo', add_unigo, methods=['POST'])

        app.add_url_rule('/del/unigo/<taxid>', 'del_unigo', del_unigo, methods=['DELETE'])

        app.add_url_rule('/build/vectors', 'build_vectors', build_vectors)

        app.add_url_rule('/list/<elemType>', 'list_elements', list_elements)

        return app

def list_elements(elemType):
    if elemType == 'vectors':
        _ = listVectors()
        print(_)
        return jsonify({ 'vectors':  _ })
      
    elif elemType == 'trees':
        return jsonify({ 'trees': listTrees() })
    
    elif elemType == 'culled':
        return jsonify({ 'culled': listCulled() })


    print(f"Unknwon element type {elemType} to list")
    abort(404)

# CPU intensive route -> offloading it to multiprocess
# using semaphore to prevent concurrency popup on multiple requests
def build_vectors():
    if _MAIN_:
        missUniversalVector = listMissUniversalVector()
        if not missUniversalVector:
            return jsonify({"status": "nothing to build"}), 200

        # global bSemaphore exists
        if bSemaphore.acquire(block=False):
            print(f"bSemaphore is acquired")
            p = Process(
                target=_buildUniversalVector,
                args=(bSemaphore, C_TYPE),
                daemon=True
            )
            p.start()
            return jsonify({
                "status": "starting",
                "targets" : missUniversalVector
            }), 200
        
        print(f"bSemaphore is locked")   
        return jsonify({"status": "running", "targets" : missUniversalVector}), 202
    else: 
        print("##### Twilight zone ######") 


@decorator
def semHolder(fn, _bSemaphore, cacheType, *args, **kwargs):
    #print("Decrorator start")
    setCacheType(cacheType)
    #time.sleep(10)
    fn(*args, **kwargs)
    _bSemaphore.release()
    #print(f"Releasing bSemaphore")

@decorator
def nsHumanizer(fn, *args, **kwargs):
    """ request.json wraps dict index by first translating its NS one-key in long description """
    _ = fn(*args, **kwargs)
    return jsonify( {
            humanReadableNS[k] : v for k, v in _.items()            
        } )

@semHolder
def _buildUniversalVector(*args,**kwargs):
    buildUniversalVector()

def ping():
    return "pong"

def view_taxids():
    return str( getTaxidKeys() )

@nsHumanizer
def view_unigos(taxid, *args, **kwargs):
    try:
        trees = getUniversalTrees(taxid)
        
    except KeyError as e:
        print(e)
        abort(404)
    return { ns : tree.serialize() for ns, tree in trees.items() }
##################################################
# ACCESSORS to taxid specific goterms as vectors
##################################################
# GET route, return universal vector # ADD NS in route ?
#@nsHumanizer
@nsHumanizer
def view_vectors(taxid, *args, **kwargs):
    try:
        taxidVectors = getUniversalVectors(taxid)
    except KeyError as e:
        print(e)
        abort(404)
    
    return taxidVectors

# POST route
# Presuming live culling is not CPU intensive enough to
# promote multiprocess offloading:: CHECK IT
@nsHumanizer
def view_culled_vectors(taxid, *args, **kwargs): 
    try:
        taxidVectors = getUniversalVectors(taxid)
    except KeyError as e:
        print(e)
        abort(404)
    try:
        data = request.get_json()
        _goParameterValidator = goParameterValidator()
        goParameter = _goParameterValidator.load(request)
        cmin, cmax, fmax = ( goParameter['minCount'],\
            goParameter['maxCount'], goParameter['maxFreq'])
    except Exception as e:
        print(e)
        print(f"Malformed GO parameters\n=>{data}")
        abort(400)
    print("Searching for culled vector...")
    try:
        culledVectors = getCulledVectors(taxid, cmin, cmax, fmax)
        print(f"Culled vector triplet found ! [{taxid} {cmin} {cmax} {fmax}]")
    except KeyError : # Culled vector not in cache, build&store
        print(f"Building culled vectors... [{taxid} {cmin} {cmax} {fmax}]")
        culledVectors = {}
        for tree_ns_key, taxidVector in taxidVectors.items():
            _ = { # Do we got it all ? Dump it once
                'registry' : taxidVector['registry'],
                'terms' : {  goID : goTerm for goID, goTerm in taxidVector['terms'].items()\
                                                        if len(goTerm['elements']) >= cmin and\
                                                           len(goTerm['elements']) <= cmax and\
                                                           goTerm['freq']          <= fmax \
                }
            }            
            storeCulledVector(_, taxid, tree_ns_key, cmin, cmax, fmax)
            culledVectors[tree_ns_key] = _

    return culledVectors



def add_unigo3NS(taxid):
    """ Add 3NS unigo object through client API
        Parameters:
        ---------- 
        taxid: ncbi taxid 
        json payload: { NS : Univgo.serialize(), }
    """
    print(f"Adding unigo3NS for taxid: {taxid}")
    _ = request.get_json()
    for tree in _.values():
        try:
            univGoTree = loadUnivGO(tree)
            storeTreeByTaxid(univGoTree, taxid)            
        except KeyError as e:
            print(f"Similar tree key already exist in database or malformed request, reject insertion\n=>{e}")
            abort(403)
        except Exception as e:
            print(f"add unigo internal error:{e}")
            abort(500)
    return jsonify({ taxid + " 3NS" : "insertion OK"})

def del_taxonymy(taxid):
    _ = deleteTaxids([taxid]) #delTreeByTaxids([taxid])
    if _ is None: 
        print(f"{taxid} not found in database, nothing to delete")       
        abort(404) 
    
    return jsonify({f"{taxid}" : { "deletions [Tree, Vector, Culled]" : list(_) }             
            })

# ADD Subroutes to delete culled and/or vector whilre preserving Trees ?

## MAY BE DECLASSIFIED
def add_unigo(taxid):
    """ Add a unigo object through client API
        Parameters:
        ---------- 
        taxid: ncbi taxid 
        json payload: ..Univgo.serialize() dict
    """
    try:
        univGoTree = loadUnivGO(request.get_json())
        storeTreeByTaxid(univGoTree, taxid)
        return jsonify({"taxid" : "insertion OK"})
    except KeyError as e:
        print(f"{taxid} already exist in database or malformed request, reject insertion\n=>{e}")
        abort(403)
    except Exception as e:
        print(f"add unigo internal error:{e}")
        abort(500)

def del_unigo(taxid):
    try:
        delTreeByTaxids([taxid])
    except KeyError:
        print(f"{taxid} not found in database, nothing to delete")       
        abort(404) 
    
    print(f"deleting related vectors...")
    delVectorsByTaxid([taxid])

    return jsonify({"taxid" : "deletion OK"})




from pyrediscore import connect, store, get, delete, listKey, setDatabaseParameters
from ...data_objects import loadUnivGO
from ....tree import enumNSkeys, assertAndCoherceValidNamespace
from decorator import decorator


######## SETTER ##########
@connect
@store
def storeTreeByTaxid(tree, taxid, *args, **kwargs):
    ns =  enumNSkeys[tree.ns]
    return  f"tree:{taxid}:{ns}", tree.serialize()

@connect
@store
def storeVector(tree, taxid=None, key=None, *args, **kwargs):
    if key:
        return  f"vector:{key}", tree.vectorize()
    if taxid:
        ns =  enumNSkeys[tree.ns]
    return  f"vector:{taxid}:{ns}", tree.vectorize()

    raise ValueError(f"No Key can be deduced for {tree}")

def storeVectorByTaxid(tree, taxid, *args, **kwargs):
    _ = storeVector(tree, taxid=taxid, *args, **kwargs)
    return _

@connect
@store
def storeCulledVector(vector, taxid, ns, cmin, cmax, fmax, *args, **kwargs):
    #print(f"##{ns}")
    #ns =  enumNSkeys[ vector["ns"] ]
    return  f"_vector:{taxid}:{ns}:{cmin}:{cmax}:{fmax}", vector


@decorator
def setNamespace(fn, *args, ns=None, **kwargs):
    _ns = list(enumNSkeys.values())
    if ns:
        _ = assertAndCoherceValidNamespace(ns)
        _ns = [_]
    return fn(*args, ns=_ns, **kwargs)

######## DELETE ##########
# These operations will complete but may raise key error if at least one entry is missing
@connect
@delete
@setNamespace
def delTreeByTaxids(taxids, *args, ns=None, **kwargs):
    """ Delete the 3 NS universal trees of the provided taxids 
    option: ns parameter can be used to delete a single namespace tree
    """
    
    return [f"tree:{taxid}:{currNS}" for currNS in ns for taxid in taxids]

@connect
@delete
@setNamespace
def delVectorByTaxids(taxids, *args, ns=None, **kwargs):
    """ Delete the 3 NS universal vectors of the provided taxids 
    option: ns parameter can be used to delete a single namespace tree
    """

    return [f"vector:{taxid}:{currNS}" for currNS in ns for taxid in taxids]

@connect
@delete # No @setNameSpace as enumeration of culled requires list operation
def delCulledVectorByTaxids(taxids, *args, ns=None, **kwargs):
    """ Delete the all culled vectors under provided taxids 
    option: ns parameter can be used to delete the set of culled vectors under
    this namespace
    """
    l = []
    for t in taxids:
        l += listCulledVectorKey(taxid=t, ns=ns)
    return [f"_vector:{_}" for _ in l]

@connect
@delete
def delVectorByKeys(keys, *args, **kwargs):
    return [f"vector:{key}" for key in keys]

@connect
@delete
def delTreeByKeys(keys, *args, **kwargs):
    return [f"tree:{key}" for key in keys]

@connect
@delete
def delCulledVectorByKeys(keys, *args, **kwargs):
    return [f"_vector:{key}" for key in keys]

def deleteTaxids(taxids, *args, **kwargs):
    deletedT = delTreeByTaxids(taxids, ignore=True)
    deletedV = delVectorByTaxids(taxids, ignore=True)
    deletedC = delCulledVectorByTaxids(taxids, ignore=True)
    if deletedC == 0 and deletedV == 0 and deletedT == 0:
        return None
    return (deletedT, deletedV, deletedC)
######## GETTER ##########

@setNamespace
def getUniversalTree3NS(taxid, *args, ns=None, raw=False, **kwargs):
    return { _ns : getUniversalTree(f"{taxid}:{_ns}", *args, raw=raw, **kwargs)\
              for _ns in ns }
        
@connect
@get
def getUniversalTree(fullKey, *args, raw=False, **kwargs):   
    return (f"tree:{fullKey}", loadUnivGO)
   
@setNamespace
def getUniversalVector3NS(taxid, *args, ns=None, raw=False, **kwargs):
    return { _ns: getUniversalVector(taxid, _ns, *args, raw=raw, **kwargs)\
              for _ns in ns }
    
@connect
@get
def getUniversalVector(taxid, ns, *args, raw=False, **kwargs):
    return (f"vector:{taxid}:{ns}", None)

@setNamespace
def getCulledVector3NS(taxid, cmin, cmax, fmax, *args, ns=None, raw=False, **kwargs):
    return {_ns : getCulledVector(taxid, _ns, cmin, cmax, fmax, *args, raw=False, **kwargs)\
              for _ns in ns }
    #return  (f"_vector:{taxid}:{cmin}:{cmax}:{fmax}", None)

@connect
@get
def getCulledVector(taxid, ns, cmin, cmax, fmax, *args, raw=False, **kwargs):    
    return (f"_vector:{taxid}:{ns}:{cmin}:{cmax}:{fmax}", None)


######## LIST ##########
@decorator
def expandQueryList(fn, *args, **kwargs):
    #print(kwargs)
    query = [ kwargs['taxid'] if not kwargs['taxid'] is None else  '*',\
              kwargs['ns']    if not kwargs['ns'] is None else  '*'
    ]
    if 'cmin' in kwargs: # One test to assume it is a culled related search
        query += [
              kwargs['cmin']  if not kwargs['cmin'] is None else  '*',\
              kwargs['cmax']  if not kwargs['cmax'] is None else  '*',\
              kwargs['fmax']  if not kwargs['fmax'] is None else  '*',\
        ]
    
    _ = ':'.join(query) 
    return fn(_, *args, **kwargs)

@connect
@listKey
@expandQueryList
def listTreeKey(*args, taxid=None, ns=None, prefix=False, **kwargs):
    #return ('tree:*', 'tree:')
    #print(f"LIST TREE KEY PREFIX {args[0]}")
    return (f"tree:{args[0]}", 'tree:')

@connect
@listKey
@expandQueryList
def listVectorKey(*args, taxid=None, ns=None, prefix=False, **kwargs):
    #return ('vector:*', 'vector:')
    #print(f"LIST VECTOR KEY PREFIX {args[0]}")
    return (f"vector:{args[0]}", 'vector:')

@connect
@listKey
@expandQueryList
def listCulledVectorKey(*args, taxid=None, ns=None, cmin=None, cmax=None, fmax=None, prefix=False, **kwargs):
    #print(f"LIST CULLED KEY PREFIX {args[0]}")
    #return ('_vector:*', '_vector:')
    return(f"_vector:{args[0]}", '_vector:')

######### CLEANUP #########
# LIST base cleanup, deletion operation can't fail
def clearTrees(*args, **kwargs):
    keyList = [ _ for _ in listTreeKey() ]
    print(f"Clearing following tree elements content:\n\t{keyList}")
    delTreeByKeys(keyList)
    return len(keyList)

def clearVectors(*args, **kwargs):
    keyList = [ _ for _ in listVectorKey() ]
    print(f"Clearing following vector elements content:\n\t{keyList}")     
    delVectorByKeys(keyList)
    return len(keyList)

def clearCulledVectors(*args, **kwargs):
    keyList = [ _ for _ in listCulledVectorKey() ]
    print(f"Clearing following culled vectors elements content:\n\t{keyList}")
    delCulledVectorByKeys(keyList)
    return len(keyList)
    
def clear(*args, **kwargs):
    print("Clearing Entiere database content")
    _ = clearTrees(*args, **kwargs)
    print(f"{_} tree(s) cleared")
    _ = clearVectors(*args, **kwargs)
    print(f"{_} vector(s) cleared")
    _ = clearCulledVectors(*args, **kwargs)
    print(f"{_} culled vector(s) cleared")

"""
scan 0 MATCH *11*
"""

"""
def delTreeByTaxids(taxids):
    miss = []
    print(f"delete redis {taxids}")
    r = redis.Redis(host=HOST, port=PORT, db=0)       
    for taxid in taxids:
        _ = r.delete(f"tree:{taxid}")
        print(f"delete redis status {str(_)}")
        if int(str(_)) != 1:
            miss.append(taxid)
    return miss:
    
def storeTreeByTaxid(tree, taxid):
    print(f"redis taxid storage adding {taxid} {tree}")
    r = redis.Redis(host=HOST, port=PORT, db=0)
    if r.exists(f"tree:{taxid}"):
        raise KeyError(f"StoreTree error: taxid {taxid} already exists in store")
    d = tree.serialize()
    r.set(taxid, json.dumps(d))

def storeVectorByTaxid(vector, taxid):
    print(f"redis taxid storage adding {taxid} {vectors}")
    r = redis.Redis(host=HOST, port=PORT, db=0)
    if r.exists(f"vector:{taxid}"):
        raise KeyError(f"StoreVector error: taxid {taxid} already exists in store")
    
    r.set(taxid, json.dumps(vector))

def getUniversalTree(taxid, raw=False):
    print(f"redis taxid storage getting {taxid}")
    r = redis.Redis(host=HOST, port=PORT, db=0)
    _ = r.get(f"tree:{taxid}")
    if not _:
        raise KeyError(f"No taxid {taxid} found in tree store")
    # _ is bytes
    return _ \
        if raw \
        else loadUnivGO( json.loads(_) )

"""


from ...api.store.client import unigoList
from prompt_toolkit import print_formatted_text, HTML
from .connect import bConnect
from . import signatureCheck
from pprint import pformat, pprint
from prompt_toolkit.styles import Style


@bConnect
@signatureCheck
def clist(*args):
    d = unigoList(*args)
    
    # The style sheet.
    """
    _style = {
        ':F': '#ff0066',
        ':P': '#ff0066',
        ':C': '#ff0066',
    }
    """
    #style =  Style.from_dict( _style )

    print_formatted_text(HTML("------ <b>Database content</b> --------".center(80)) )
    _ = pformat(d, depth=4, sort_dicts=True, width=80)
    print_formatted_text(_)
    # Print stylinf


import requests, time

HOSTNAME=None
PORT=None

class ClientError(Exception):
    def __init__(self, url):
        self.url = url

class BuildConnectionError(ClientError):
    def __init__(self, url, code):
        super().__init__(url)
        self.code = code
    def __str__(self):
        return f"Error {self.code} while monitoring build process [{self.url}]"

class InsertionError(ClientError):
    def __init__(self, url, code):
        super().__init__(url)
        self.code = code
    def __str__(self):
        return f"Insertion was denied, The trees may already exist in database [{self.url}]"

class DeletionError(ClientError):
    def __init__(self, url, code):
        super().__init__(url)
        self.code = code
    def __str__(self):
        return f"Deletion was denied, The trees may not exist in database [{self.url}]"

def handshake(hostname, port):
    try:
        url = f"http://{hostname}:{port}/ping"
        req = requests.get(url)
    except ConnectionError as e:
        raise ConnectionError(f"Unable to handshake at {url}\n->{e}")

    if not req.status_code == requests.codes.ok:
        raise ConnectionError(f"Error {req.status_code} while handshaking at {url}")
        if not req.text == "pong":
            raise ConnectionError(f"Improper handshake ({req.text}) at {url}")
    
    global HOSTNAME, PORT
    
    HOSTNAME = hostname 
    PORT     = port
    return True

def addTree3NSByTaxid(treeTaxidIter, fromCli=False):
    
    requestedTree = {}
    for taxid, _, tree in treeTaxidIter:       
        requestedTree[ f"{taxid}:{tree.ns}" ] = tree.serialize()
        
    url = f"http://{HOSTNAME}:{PORT}/add/taxid/{taxid}"
    req = requests.post(url, json=requestedTree)
    
    if req.status_code == requests.codes.ok:
        msg = f"Successfull tree adding at {url}"
        if fromCli:
            return msg
        print(msg)
        
    else:
        if fromCli:
            raise InsertionError(url, req.status_code)
        print(f"Error {req.status_code} while inserting at {url}") 

def delTaxonomy(taxids, fromCli=False):
    #print(f"Want to del by taxids {taxids}")
    for taxid in taxids:
        url = f"http://{HOSTNAME}:{PORT}/del/taxid/{taxid}"
        req = requests.delete(url)

        if req.status_code == requests.codes.ok:
            msg = f"Successfully deleted data under taxonomy {taxid} [{url}]"
            if fromCli:
                return msg
            print(msg)
                
        else:
            if fromCli:
                raise DeletionError(url, req.status_code)            
            print(f"Error {req.status_code} while deleting at {url}")


def buildVectors(fromCli=False):
    """ Trigger Vector building
        * List total number of vector to build
        * regularly asks for status
    """
    (status, size) = _pingAndUnwrapBuildReq(fromCli)
    yield (status, size)
    while status in ["starting", "running"]:
        time.sleep(1)
        (status, _size) = _pingAndUnwrapBuildReq(fromCli)
        if _size < size:
            for iSize in range(size, _size, -1):
                time.sleep(0.5)
                if iSize - 1 > 0:
                    yield (status, iSize - 1)
            size = _size
    yield ("completed", 0)


def _pingAndUnwrapBuildReq(fromCli):
    url = f"http://{HOSTNAME}:{PORT}/build/vectors"
    req = requests.get(url)
    if not req.status_code in ["200", "202"]:
        data = req.json()
        status = data["status"]
        n = 0 if status == "nothing to build" else len(data["targets"])
        return (status, n)
    else:
        if fromCli:
            raise BuildConnectionError(url, req.status_code)            
        print(f"Error {req.status_code} while buidling at {url}")



def unigoList(_elem="all"):
    
    d = {}
    for elem in ("trees", "vectors", "culled"):
        if _elem == "all" or elem  == _elem :
            url = f"http://{HOSTNAME}:{PORT}/list/{elem}"
            req = requests.get(url)
            d[elem] = req.json()[elem]
    return d


from flask import Flask, abort, jsonify, request
import requests, json
from marshmallow import EXCLUDE
from .. import vloads as createGOTreeTestUniverseFromAPI
from .. import uloads as createGOTreeTestFromAPI
from .. import utils
from .io import checkPwasInput
from ..stat_utils import applyOraToVector, kappaClustering
from .data_objects import CulledGoParametersSchema as goParameterValidator
from copy import deepcopy as copy

GOPORT=1234
GOHOST="127.0.0.1"
def listen(goApiHost:str, goApiPort:int, vectorized:bool):
    global GOPORT, GOHOST
    GOPORT = goApiPort
    GOHOST = goApiHost

    app = Flask(__name__)
    app.config['JSON_SORT_KEYS'] = False # To keep dict order in json
    app.add_url_rule("/", 'hello', hello)
    if vectorized:
        print(f"PWAS API vector listening")
        app.add_url_rule("/compute", "computeOverVector", computeOverVector, methods=["POST"])
    else:
        print(f"PWAS API tree listening")
        app.add_url_rule("/compute", "computeOverTree", computeOverTree, methods=["POST"])
    
    app.add_url_rule("/loadVector/<taxid>", loadVector) # WARNING : don't work
    return app

def hello():
    return "Hello pwas"

def computeOverVector():
    forceUniversal = False
    data = checkPwasInput()
   
    print(f'I get data with {len(data["all_accessions"])} proteins accessions including {len(data["significative_accessions"])} significatives')
    
    if forceUniversal:
        go_resp = utils.unigo_vector_from_api(GOHOST, GOPORT, data["taxid"])
    else:
         # Culling vector parameters
        _goParameterValidator = goParameterValidator()
        goParameter = _goParameterValidator.load(request,  unknown=EXCLUDE)
        go_resp = utils.unigo_culled_from_api(GOHOST, GOPORT, data["taxid"], goParameter)

    if go_resp.status_code != 200:
        print(f"ERROR request returned {go_resp.status_code}")
        abort(go_resp.status_code)
    
    # Here Contract of 3 NS on go_resp
    vectorizedProteomeTrees = go_resp.json()

    kappaClusters = kappaClusteringOverNS(vectorizedProteomeTrees, data)
    return jsonify(kappaClusters)

def fuseVectorNameSpace(_vectorElements, merge):
    vectorElements = copy(_vectorElements)

    if not merge:
        for ns, vectorElement in vectorElements.items():
            for goID, goVal in vectorElement["terms"].items():
                goVal["ns"] = ns
        return vectorElements

    fusedNS = {
        "terms" : {},
        "registry": []
    }

    for ns, vectorElement in vectorElements.items():
        # All registries are identical
        if not fusedNS["registry"]:
            fusedNS["registry"] = vectorElement["registry"]
        for goID, goVal in vectorElement["terms"].items():
            goVal["ns"] = ns
            fusedNS["terms"][goID]  = goVal
    return { "fusedNS" : fusedNS }

def kappaClusteringOverNS(_vectorElements, expData, merge=False):

    vectorElements = fuseVectorNameSpace(_vectorElements, merge)  
    kappaClusters = {}   

    #print("expData", expData)
    if expData.get("pvalue"):
        pvalue = expData["pvalue"]
    else:
        pvalue = 0.05

    print("pvalue", pvalue)

    for ns, vectorElement in vectorElements.items():
        res = applyOraToVector(vectorElement,\
            expData["all_accessions"],\
            expData["significative_accessions"],\
            pvalue)
        formatted_res = [{**{"go": go_term}, **res[go_term]} for go_term in res]
        if len( res.keys() ) <= 1:
            kappaClusters[ns] = {'Z': res, 'list' : formatted_res}
        else:
            Z = kappaClustering(vectorElement["registry"], res)
            kappaClusters[ns] = {'Z': Z, 'list' : formatted_res}
    
    return(kappaClusters)

def computeOverTree():
    data = checkPwasInput() 
    print(f'I get data with {len(data["all_accessions"])} proteins accessions including {len(data["significative_accessions"])} significatives')

    go_resp = utils.unigo_tree_from_api(GOHOST, GOPORT, data["taxid"])

    if go_resp.status_code != 200:
        print(f"ERROR request returned {go_resp.status_code}")
        abort(go_resp.status_code)
    
    print("Create Unigo tree")
    unigoTreeFromAPI = createGOTreeTestFromAPI(go_resp.text, data["all_accessions"])
    x,y = unigoTreeFromAPI.dimensions
    print("Unigo Object successfully buildt w/ following dimensions:")
    print(f"\txpTree => nodes:{x[0]} children_links:{x[1]}, total_protein_occurences:{x[2]}, protein_set:{x[3]}")  
    print(f"\t universeTree => nodes:{y[0]} children_links:{y[1]}, total_protein_occurences:{y[2]}, protein_set:{y[3]}")  

    #Compute fisher stats
    if data["method"] == "fisher":
        print("Computing ORA with fisher")
        rankingsORA = unigoTreeFromAPI.computeORA(data["significative_accessions"], verbose = False)
        
        return rankingsORA.json

    return {"not computed": "unavailable stat method"}

def _loadVector(taxid):
    go_resp = utils.unigo_vector_from_api(GOPORT, taxid)
    if go_resp.status_code != 200:
        print(f"ERROR request returned {go_resp.status_code}")
        abort(go_resp.status_code)
    else:
        return go_resp

def loadVector(taxid):
    if _loadVector(taxid):
        return {"ok" : f"Vector loaded for {taxid} taxid"}
    else:
        return {"error" : f"Can't load vector for {taxid} taxid"}







from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.auto_suggest import AutoSuggest, Suggestion
from decorator import decorator

import re
"""
 CLI command Signature declaration and check
 we check command and subcommand

 To be able to check parameters, we could declare

 declare = {
     'connect' : { [""]
 }
"""

signatures = {
    'clist': { 
        'vector',
        'tree',
        'culled',
        'all'
    },
    'connect' : None,    
    'build'   : None,
    'load'    : None,
    'exit'    : None,
    'delete'  : None,
    'help'    : {'clist', 'connect',  'load', 'exit', 'delete', 'build'}

}

completer = NestedCompleter.from_nested_dict(signatures)

class SignatureBaseError(Exception):
    def __init__(self, cmd):
        self.cmd = cmd
        self.availbleCmd = set(signatures.keys())
        if cmd in self.availbleCmd:
            self.subCmd = signatures[cmd] if type(signatures[cmd]) == set else set(signatures[cmd].keys())
        super().__init__()
        
class SignatureCallError(SignatureBaseError):
    def __init__(self, cmd):
        super().__init__(cmd)#self.message)
    def __str__(self):
        return f"<ansired><u>{self.cmd}</u> is not a valid command call </ansired> <ansigreen><i>{self.availbleCmd}</i><ansigreen>"

class SignatureWrongSubCommandError(SignatureBaseError):
    def __init__(self, cmd, *subCmd):
        super().__init__(cmd)#self.message)
        self.badSubCmd = subCmd
    def __str__(self):
        return f"<ansired>{self.cmd} was called with wrong subcommand <u>{self.badSubCmd}</u> instead of <i>{self.subCmd}</i></ansired>"

class SignatureEmptySubCommandError(SignatureBaseError):
    def __init__(self, cmd):
        super().__init__(cmd)#self.message)
    def __str__(self):
        return f"<ansired><u>{self.cmd}</u> empty argument list instead of <i>{self.subCmd}</i></ansired>"

# Should be made recursive if needed
@decorator
def signatureCheck(fn, *args, **kwargs):
    cmd = fn.__name__
    if not cmd in signatures:
        raise SignatureCallError(cmd)
    if signatures[cmd] is None:
        return fn(*args, **kwargs)
    availArg = set([ subCmd for subCmd in signatures[cmd] ])
    
    _ = set(args) - set(availArg)
    if not args:
        raise SignatureEmptySubCommandError(cmd)        
    if _:
        raise SignatureWrongSubCommandError(cmd, *_)        
    return fn(*args, **kwargs)

class customAutoSuggest(AutoSuggest):
    def __init__(self):
        super().__init__()
    def update(self, cmd):
        self.currentCommand = cmd

    def get_suggestion(self, buffer, document):
        if str(document.text).startswith('connect'):
            _ = re.findall('([\S]+)', document.text)
            if len(_) == 1:
                return Suggestion(" 127.0.0.1 1234")
            if len(_) == 2:
                return Suggestion(" 1234")
            
        #if str(document.text).startswith('list'):
        #    return Suggestion(" all|vector|culled|tree")
        if str(document.text).startswith('load'):
            _ = re.findall('([\S]+)', document.text)
            if len(_) > 1:
                return Suggestion( pathSuggest(_[-1]) )

        return Suggestion("")

import glob 
from os.path import commonprefix
def pathSuggest(string):
    _ = glob.glob(f"{string}*")
    return re.sub(rf'^{string}', '',commonprefix(_) )



from . import redis
from . import local

CACHE_PKG=local
T_CACHE_TYPE=["local", "redis"]
CACHE_SYMBOL="local"

def setCacheType(_type, **kwargs):
    if _type not in T_CACHE_TYPE:
        raise ValueError(f"{_type} is not registred cache type {T_CACHE_TYPE}")
    global CACHE_PKG, CACHE_SYMBOL
    CACHE_PKG = redis if _type == "redis" else local
    CACHE_SYMBOL = _type
    #print(f"Set cache to {_type}")

    if _type == 'redis':
        p = {}
        if 'rp' in kwargs:
            p['port'] = kwargs['rp']
        if 'rh' in kwargs:
            p['host'] = kwargs['rh']
        
        CACHE_PKG.setDatabaseParameters(**p)

def storeTreeByTaxid(tree, taxid):
    return CACHE_PKG.storeTreeByTaxid(tree, taxid)

def delTreeByTaxids(taxids):
    return CACHE_PKG.delTreeByTaxids(taxids)

def delVectorsByTaxid(taxids, delAllRelated=True):
    CACHE_PKG.delVectorsByTaxid(taxids)
    if CACHE_SYMBOL == 'redis' and delAllRelated:
        CACHE_PKG.delCulledByTaxids(taxids)

def getTaxidKeys():
    return CACHE_PKG.getTaxidKeys()

def getUniversalTrees(taxid, raw=False):
    # if not local deserialize or deeper

    return CACHE_PKG.getUniversalTree3NS(taxid, raw=raw)

def getUniversalVectors(taxid):
    try:
        vectors = CACHE_PKG.getUniversalVector3NS(taxid)
        print(f"{taxid} Universal vector in cache")
    except KeyError:
        try:
            tree3NS = CACHE_PKG.getUniversalTree3NS(taxid)
        except KeyError:
            raise KeyError(f"Vector error, No taxid {taxid} in stores")

        print(f"Building 3NS {taxid} Universal vector")
        vectors = {}
        for tree in tree3NS:
            _ = tree.vectorize()
            vectors[ tree.ns ] = _
            CACHE_PKG.storeVectorByTaxid(_, taxid)
    
    return vectors

### FOLLOWING TWO SHOULD BE REMOVED ###
def _getUniversalTree(taxid, raw=False):
    # if not local deserialize or deeper

    return CACHE_PKG.getUniversalTree(taxid, raw=raw)

def _getUniversalVector(taxid):
    try:
        vec = CACHE_PKG.getUniversalVector(taxid)
        print(f"{taxid} Universal vector in cache")
    except KeyError:
        try:
            tree = CACHE_PKG.getUniversalTree(taxid)
        except KeyError:
            raise KeyError(f"Vector error, No taxid {taxid} in stores")

        print(f"Building {taxid} Universal vector")
        vec = tree.vectorize()
        CACHE_PKG.storeVectorByTaxid(vec, taxid)
    
    return vec

def clear():
    CACHE_PKG.clear()

def deleteTaxids(taxids):
    return CACHE_PKG.deleteTaxids(taxids)

def status():
    if CACHE_SYMBOL == 'redis':
        return len(listTrees()), len(listVectors()), len(listCulled())

def listTrees():
    if CACHE_SYMBOL == 'redis':
        return [ _ for _ in CACHE_PKG.listTreeKey(prefix=False) ]

def listVectors():
    if CACHE_SYMBOL == 'redis':
        return [ _ for _ in CACHE_PKG.listVectorKey(prefix=False) ]
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL KEYS ITER")

def listCulled():
    if CACHE_SYMBOL == 'redis':
        return [ _ for _ in CACHE_PKG.listCulledVectorKey(prefix=False) ]
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL KEYS ITER")


def getCulledVectors(taxid, cmin, cmax, fmax):
    if CACHE_SYMBOL == 'redis':
        _ = CACHE_PKG.getCulledVector3NS(taxid, cmin, cmax, fmax)      
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL getCulledVector")
    return _

def storeCulledVector(vector, taxid, ns, cmin, cmax, fmax):
    if CACHE_SYMBOL == 'redis':
        _ = CACHE_PKG.storeCulledVector(vector, taxid, ns, cmin, cmax, fmax)
    else:
        raise TypeError("YOU SHOULD IMPLEMENT LOCAL storeCulledVector")
    return _   

def listMissUniversalVector():
    _treeID   = set( listTrees() )
    _vectorID = set( listVectors() )
    #print(f"{_treeID}  -----  {_vectorID}")
    return list(_treeID - _vectorID)

def buildUniversalVector():
    taxid_ns_treeKeys = listMissUniversalVector()
    for taxid_ns_treeKey in taxid_ns_treeKeys:
        tree = CACHE_PKG.getUniversalTree(taxid_ns_treeKey)
        print(f"Build vector for {taxid_ns_treeKey} from {tree}")
        CACHE_PKG.storeVector(tree, key=taxid_ns_treeKey)



from decorator import decorator
from prompt_toolkit import print_formatted_text, HTML

from ..helpers import signatureCheck, signatureCheck
from .mutators import load, delete, build
from .viewers import clist
from .connect import connect, getPrompt
from decorator import decorator
"""
class ExecutorBaseError(Exception):
    def __init__(self, symbol):
        self.typings = cmd
        self.availbleCmd = set(signatures.keys())
        if cmd in availbleCmd:
            self.subCmd = signatures[cmd] if type(signatures[cmd]) == set else set(signatures[cmd].keys())
        super().__init__()
        
class ExecutorParameterNumberError(SignatureBaseError):
    def __init__(self, cmd):
        super().__init__(cmd)#self.message)
    def __str__(self):
        return f"<ansired><u>{self.cmd}</u> is not a valid command call </ansired> <ansigreen><i>{self.availbleCmd}</i><ansigreen>"

class ExecutorParameterNumberError(SignatureBaseError):
    def __init__(self, cmd):
        super().__init__(cmd)#self.message)
    def __str__(self):
        return f"<ansired><u>{self.cmd}</u> is not a valid command call </ansired> <ansigreen><i>{self.availbleCmd}</i><ansigreen>"
"""



class Executor():
    def __init__(self):
        self.executor = {
            "clist" : {
                "_target" : clist,
                "paramTypes" : None,
                "help" : f"List database content\nUsage: clist tree|all|culled|vector"
            },
            "connect" : {
                "_target"    : connect,
                "paramTypes" : [str, int],
                "runningParams" : True,
                "help" : f"Connect to database\nUsage: connect <i>hostname</i> <i>port</i>"
            },
            "load"    :  {
                "_target"    : load,
                "paramTypes" : [str, str],
                "runningParams" : True,
                "help" : f"Insert any number of triplets of GO trees in the database\nUsage: load <i>owl_file proteome_file_1, ...</i>"
            },
            'exit'    :  {
                "_target"    : _exit,
                "paramTypes" : None,
                "help" : f"Close the interface\nUsage: exit"
            },
            "delete"    :  {
                "_target"    : delete,
                "paramTypes" : [str],
                "runningParams" : True,
                "help" : f"Delete any number of triplets of GO trees in the database\nUsage: delete <i>taxid_1, ...</i>"
            },
            "build"    :  {
                "_target"    : build,
                "paramTypes" : None,               
                "help" : f"Build all missing triplets of <u>GO vectors</u> in the database\nUsage: build"
            }
        }
    
    @property
    def availbleCmd(self):
        for k in self.executor:
            yield k
        
    def isa(self, cmd):
        return cmd in self.executor
    
    def help(self, cmd):
        return self.executor[cmd]["help"]

    def process(self, cmd, *args):
        if not self.isa(cmd):
            raise ValueError(f"{cmd} is not a valid command {self.executor.keys()}")
        fn = self.executor[cmd]["_target"]
        try:
            self.checkTypes(cmd, *args)
            fn(*args)
        except Exception as e:
            print_formatted_text(HTML(e))
            return False
        return True

    def checkTypes(self, symbol, *args):
        _ = self.executor[symbol]
        typeArgs = _["paramTypes"]
        if typeArgs is None:
            return True
        
        if len(args) < len(typeArgs) or\
            (len(args) > len(typeArgs) and _["runningParams"]) :
            raise TypeError(f"<ansired>{symbol} excepts {len(typeArgs)} argument(s)</ansired>")
""" GIVING UP ON PARAM TYPE CHECKING...
        for i, carg in enumerate(args):
            ctype=type(carg)
            j = i if i < len(typeArgs) else len(typeArgs) - 1
            _type = typeArgs[j]
            
            if ctype != _type:
                _type = str(_type).replace('<', '').replace('>', '').replace("'", "")
                raise TypeError(f"<ansired>{symbol} argument {carg} wrongs types should be {_type}</ansired>")
                #print(f"<ansired>{symbol} argument {carg} has wrong {ctype}</ansired>")
                #raise TypeError(f"<ansired>{symbol} argument {carg} has wrong {ctype}</ansired>")
"""
def _exit():
    print_formatted_text(HTML(f"\n\n<skyblue>See you space cowboy</skyblue>"))
    exit(0)

"""
def stripTaxid(data):
    ok     = []
    notok  = []
    for k in data if re.match("^[^\:]")
"""









from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.shortcuts import ProgressBar
from .connect import bConnect
from . import signatureCheck
from ...utils import loadUniversalTreesFromXML
from ...api.store.client import addTree3NSByTaxid as goStoreAdd, delTaxonomy, buildVectors
from ...api.store.client import InsertionError, DeletionError
import time

@bConnect
@signatureCheck
def load(owlFile, *args):
    print(f"Connecting to goStore service to add following proteome(s) {args}")
    taxidTreeIter = loadUniversalTreesFromXML(args, owlFile)
    try:
        msg = goStoreAdd(taxidTreeIter, fromCli=True)
        print_formatted_text(HTML(f'<ansigreen>{msg}</ansigreen>'))
    except InsertionError as e:
        print_formatted_text(HTML(f'<ansired>{e}</ansired>'))

    

@bConnect
@signatureCheck
def delete(*taxids):
    try:
        msg = delTaxonomy(taxids, fromCli=True)
    except DeletionError as e:
        print_formatted_text(HTML(f'<ansired>{e}</ansired>'))
    
    print_formatted_text(HTML(f'<ansigreen>{msg}</ansigreen>'))

@bConnect
@signatureCheck
def build():
    try:
        bVectorStatusChange = buildVectors(fromCli=True)
        status, size = next(bVectorStatusChange)
        if status == "nothing to build":
            print_formatted_text( HTML(f'<ansigreen>No vector to build</ansigreen>') )
            return

        if status == "running":
            print_formatted_text( HTML(f'<ansigreen>A building thread is already running, monitoring it...</ansigreen>') )
        else:
            print_formatted_text( HTML(f'<ansigreen>Start Building</ansigreen>')) 
        with ProgressBar() as pb:
            for i in pb( bVectorStatusChange, total=int(size) ):
                time.sleep(0.1)
    except Exception  as e:
        print_formatted_text( HTML(f'<ansired>Error while Building {e}</ansired>') )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['unigo',
 'unigo.api',
 'unigo.api.store',
 'unigo.api.store.cache',
 'unigo.repl',
 'unigo.repl.commands',
 'unigo.tree']

package_data = \
{'': ['*'], 'unigo': ['data/*']}

install_requires = \
['Flask>=2.0.2,<3.0.0',
 'Owlready2>=0.36,<0.37',
 'docopt>=0.6.2,<0.7.0',
 'marshmallow>=3.14.1,<4.0.0',
 'requests>=2.27.1,<3.0.0',
 'requires>=0.10.2,<0.11.0',
 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'unigo',
    'version': '2.0.2',
    'description': '',
    'long_description': None,
    'author': 'glaunay',
    'author_email': 'pitooon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
